package gs.datagen.v0.generators

import gs.datagen.v0.Gen
import java.time.LocalDate
import java.util.Random

/** Base for generators that produce `java.time.LocalDate` values.
  */
trait GenLocalDate extends Gen[LocalDate]

/** Provids implementations for [[GenLocalDate]].
  */
object GenLocalDate:

  /** Implementation of [[GenLocalDate]] that selects random dates before some
    * given date.
    *
    * @param pivot
    *   The pivot date before which random values are selected.
    * @param days
    *   The day bound.
    * @param months
    *   The month bound.
    * @param years
    *   The year bound.
    */
  final class Before(
    val pivot: LocalDate,
    val days: MinMax,
    val months: MinMax,
    val years: MinMax
  ) extends GenLocalDate:

    /** @inheritDocs
      */
    override def generate(input: Any): LocalDate =
      val rng = Gen.rng()
      pivot
        .minusDays(days.select(rng).toInt)
        .minusMonths(months.select(rng).toInt)
        .minusYears(years.select(rng).toInt)

  /** Implementation of [[GenLocalDate]] that selects random dates after some
    * given date.
    *
    * @param pivot
    *   The pivot date after which random values are selected.
    * @param days
    *   The day bound.
    * @param months
    *   The month bound.
    * @param years
    *   The year bound.
    */
  final class After(
    val pivot: LocalDate,
    val days: MinMax,
    val months: MinMax,
    val years: MinMax
  ) extends GenLocalDate:

    /** @inheritDocs
      */
    override def generate(input: Any): LocalDate =
      val rng = Gen.rng()
      pivot
        .plusDays(days.select(rng).toInt)
        .plusMonths(months.select(rng).toInt)
        .plusYears(years.select(rng).toInt)

  /** Implementation of [[GenLocalDate]] that selects random dates centered on
    * some given date.
    *
    * @param pivot
    *   The pivot date around which random values are selected.
    * @param days
    *   The day bound.
    * @param months
    *   The month bound.
    * @param years
    *   The year bound.
    */
  final class Around(
    val pivot: LocalDate,
    val days: MinMax,
    val months: MinMax,
    val years: MinMax
  ) extends GenLocalDate:

    /** @inheritDocs
      */
    override def generate(input: Any): LocalDate =
      val rng = Gen.rng()
      pivot
        .plusOrMinusDays(rng, days)
        .plusOrMinusMonths(rng, months)
        .plusOrMinusYears(rng, years)

  extension (base: LocalDate)

    /** Select a bounded random number of days before or after the base date.
      *
      * @param rng
      *   The random generator.
      * @param range
      *   The bounded range.
      * @return
      *   The new date.
      */
    def plusOrMinusDays(
      rng: Random,
      range: MinMax
    ): LocalDate =
      if rng.nextBoolean() then base.plusDays(range.select(rng).toInt)
      else base.minusDays(range.select(rng).toInt)

    /** Select a bounded random number of months before or after the base date.
      *
      * @param rng
      *   The random generator.
      * @param range
      *   The bounded range.
      * @return
      *   The new date.
      */
    def plusOrMinusMonths(
      rng: Random,
      range: MinMax
    ): LocalDate =
      if rng.nextBoolean() then base.plusMonths(range.select(rng).toInt)
      else base.minusMonths(range.select(rng).toInt)

    /** Select a bounded random number of years before or after the base date.
      *
      * @param rng
      *   The random generator.
      * @param range
      *   The bounded range.
      * @return
      *   The new date.
      */
    def plusOrMinusYears(
      rng: Random,
      range: MinMax
    ): LocalDate =
      if rng.nextBoolean() then base.plusYears(range.select(rng).toInt)
      else base.minusYears(range.select(rng).toInt)

end GenLocalDate
